from abc import ABC, ABCMeta, abstractmethod
from typing import List, Optional, Dict

from .algorithm import detect_involved_tables
from .scope import QueryScope, ProjectionScope, QueryNodeParameter, FunctionCallScope
from .select import Select
from .sql import SqlNode, SqlContext, SqlPredicate, SqlOrderByItem, SqlLimit, SqlOffset, \
    SqlKeyword, SqlAggFunctionCall, K_INNER_JOIN, SqlFunctionCall, SqlSumFunctionCall, \
    K_LEFT_JOIN, Subquery, SqlComparison, SqlColumn, O_EQ, SqlContextWrapper, SqlLogicalExpression, OverridableSqlNode, \
    K_CROSS_JOIN, SqlIdentifier, SqlTable, SqlAdditiveAggrFunctionCall


class SqlPredicateAppender(metaclass=ABCMeta):

    @abstractmethod
    def append(self, query_scope: QueryScope, table: SqlTable): pass


class QueryNode(SqlNode, ABC):

    # noinspection PyTypeChecker
    def __init__(self, query_parameter: QueryNodeParameter):
        self.query_parameter = query_parameter
        self.left: QueryNode = None
        self.right: QueryNode = None
        self._children_joiner = K_INNER_JOIN
        self.sql_node_reference: Dict[SqlNode, SqlNode] = {}

    def to_sql(self, context: 'SqlContext') -> str:
        return self._create_query(context).to_sql(context)

    def _create_query(self, context: 'SqlContext') -> Select: pass

    def get_reference(self, sql_node: SqlNode) -> SqlNode:
        return self.sql_node_reference.get(sql_node)

    def has_reference(self, sql_node: SqlNode) -> bool:
        return sql_node in self.sql_node_reference


class RootableQueryNode(QueryNode, ABC):

    def __init__(self, query_parameter: QueryNodeParameter, is_root: Optional[bool] = False):
        super().__init__(query_parameter)
        self._is_root = is_root

    def _wrap_root(self, query: Select, query_node: QueryNode, projection_scope: ProjectionScope) -> Select:
        select = Select()
        for projection in projection_scope.projections:
            reference = query_node.get_reference(projection)
            column = SqlColumn(query, reference)
            select.project(self._rewrite_root_projection(projection, column))
        select.from_(Subquery(query))
        self._fill_root_query_parameters(select)
        return select

    def _fill_root_query_parameters(self, select: Select, query_parameter: Optional[QueryNodeParameter] = None):
        if query_parameter is None:
            query_parameter = self.query_parameter

        select.limit(query_parameter.limit) \
            .offset(query_parameter.offset) \
            .order_by_list(query_parameter.order_by_items)

    def _rewrite_root_projection(self, projection: SqlNode, rewritten: SqlNode):
        if isinstance(projection, OverridableSqlNode):
            from copy import deepcopy

            projection = deepcopy(projection)
            if hasattr(projection, 'original_alias') and projection.original_alias is not None:
                rewritten.alias = projection.original_alias
            projection.override(rewritten)
        return projection


class SimpleQueryNode(RootableQueryNode):

    def __init__(self, query_parameter: QueryNodeParameter, is_root: Optional[bool] = False):
        super().__init__(query_parameter)
        self._is_root = is_root

    def _create_query(self, context: 'SqlContext') -> Select:
        self.query_parameter = self._rewrite_query_parameter(context, self.query_parameter)

        query_scope = QueryScope.detect(context, [
            self.query_parameter.projections,
            self.query_parameter.predicates,
            self.query_parameter.order_by_items,
            self.query_parameter.group_bys
        ])
        select = Select()

        if query_scope.has_multiple_tables():
            self._build_join(context, query_scope, select)
        else:
            select.from_(query_scope.get_single_table())

        self._build_projections(context, select)
        self._build_predicates(select)
        self._build_group_by(select)
        self._fill_root_query_parameters(select)

        return select

    def _rewrite_query_parameter(self, context: SqlContext, query_parameter: QueryNodeParameter):
        return query_parameter

    def _build_projections(self, context: SqlContext, select: Select):
        for projection in self.query_parameter.projections:
            if hasattr(projection, "alias"):
                if not self._is_root:
                    projection.alias = context.alias(projection)
                    self.sql_node_reference[projection] = projection.alias
            select.project(projection)

    def _build_group_by(self, select: Select):
        if self.query_parameter.group_bys is not None:
            for item in self.query_parameter.group_bys:
                if isinstance(item, SqlIdentifier):
                    select.group_by(item)
                else:
                    if item.alias is not None:
                        select.group_by(item.alias)
                    else:
                        select.group_by(item)

    def _build_predicates(self, select: Select):
        if self.query_parameter.predicates is not None:
            for predicate in self.query_parameter.predicates:
                if isinstance(predicate, SqlPredicate):
                    if isinstance(predicate.operand, SqlAggFunctionCall):
                        select.having(predicate)
                    else:
                        select.predicate(predicate)
                else:
                    raise RuntimeError(f"Invalid predicate : {predicate}")

    def _build_join(self, context: SqlContext, query_scope: QueryScope, select: Select):
        join_order = list()
        queried_tables = query_scope.get_tables()
        relation_graph = context.get_relation_graph()

        queried_tables_name = sorted(list(queried_tables.keys()))
        main_table = query_scope.get_main_table(context)
        parent_tree, _ = detect_involved_tables(main_table, queried_tables_name, {main_table}, relation_graph,
                                                join_order=join_order)
        for fk_table_name in join_order:
            parent_table_name = parent_tree[fk_table_name]
            table_schema = context.get_table_schema(parent_table_name)

            if table_schema.has_join(fk_table_name):
                join = table_schema.get_join(fk_table_name)
                join_condition = context.get_join_condition(parent_table_name, join.target)
                select.join(SqlKeyword(join.join_type), queried_tables.get(join.target), join_condition)
            else:
                join = context.get_table_schema(fk_table_name).get_join(parent_table_name)
                join_condition = context.get_join_condition(join.target, fk_table_name)
                select.join(SqlKeyword(join.join_type), queried_tables.get(fk_table_name), join_condition)
        select.from_(queried_tables.get(main_table))


class QueryCombineNode(RootableQueryNode):

    def __init__(self,
                 projection_scope: ProjectionScope,
                 order_by_items: Optional[List[SqlNode]] = None,
                 limit: Optional[SqlNode] = None,
                 offset: Optional[SqlNode] = None,
                 is_root: Optional[bool] = False):
        super().__init__(QueryNodeParameter(
            order_by_items=order_by_items,
            limit=limit,
            offset=offset
        ), is_root)
        self._projection_scope = projection_scope

    def _create_query(self, context: 'SqlContext') -> Select:
        if self.right is None:
            query = self.left._create_query(context)
            if self._is_root:
                query = self._wrap_root(query, self.left, self._projection_scope)
            else:
                self.sql_node_reference = self.left.sql_node_reference
        else:
            query = self._build_binary_query(context)

        return query

    def _build_binary_query(self, context: 'SqlContext') -> Select:
        select = Select()
        left_query = self.left._create_query(context)
        right_query = self.right._create_query(context)

        select.from_(Subquery(left_query))
        self._build_projections(left_query, right_query, select, context)
        self._build_join(left_query, right_query, select)

        if self._is_root:
            self._fill_root_query_parameters(select)

        return select

    def _build_join(self, left_query: QueryNode, right_query: QueryNode, select: Select):
        if self._projection_scope.has_dimensions():
            comparisons = []
            for dimension in self._projection_scope.dimensions:
                left_column = SqlColumn(left_query, self.left.get_reference(dimension))
                right_column = SqlColumn(right_query, self.right.get_reference(dimension))
                comparisons.append(SqlComparison(left_column, O_EQ, right_column))
            select.join(K_INNER_JOIN, Subquery(right_query), SqlLogicalExpression(*comparisons))
        else:
            select.join(K_CROSS_JOIN, Subquery(right_query))

    def _build_projections(self, left_query: QueryNode, right_query: QueryNode, select: Select, context: 'SqlContext'):
        projections = self._projection_scope.projections

        for projection in projections:
            if self.left.has_reference(projection):
                reference = self.left.get_reference(projection)
                column = SqlColumn(left_query, reference)
            else:
                reference = self.right.get_reference(projection)
                column = SqlColumn(right_query, reference)

            if reference is not None:
                column.alias = context.alias(column)
                self.sql_node_reference[projection] = column.alias

                if self._is_root:
                    select.project(self._rewrite_root_projection(projection, column))
                else:
                    select.project(column)

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(SqlContextWrapper(context))


class AdditiveAggrQueryCombineNode(QueryNode):

    def __init__(self, dimensions: List[SqlNode], aggr_function_calls: List[SqlFunctionCall],
                 function_call_primary_key: SqlColumn):
        super().__init__(QueryNodeParameter())
        self._dimensions = dimensions
        self._aggr_function_calls = aggr_function_calls
        self._function_call_primary_key = function_call_primary_key
        self.children_joiner = K_LEFT_JOIN

    def _create_query(self, context: 'SqlContext') -> Select:
        select = Select()
        dimension_query = self.left._create_query(context)
        aggr_function_call_query = self.right._create_query(context)

        select.from_(Subquery(dimension_query))

        self._build_dimension_projection(context, select, dimension_query)
        self._build_aggr_function_call_projection(context, select, aggr_function_call_query)

        dimension_primary_key_column = SqlColumn(dimension_query,
                                                 self.left.get_reference(self._function_call_primary_key))
        aggr_function_call_primary_key_column = SqlColumn(aggr_function_call_query,
                                                          self.right.get_reference(self._function_call_primary_key))
        select.join(self.children_joiner, Subquery(aggr_function_call_query),
                    SqlComparison(dimension_primary_key_column, O_EQ, aggr_function_call_primary_key_column))

        return select

    def _build_dimension_projection(self, context: SqlContext, select: Select, dimension_query: Select):
        for dimension in self._dimensions:
            dimension_column = SqlColumn(dimension_query, self.left.get_reference(dimension))
            dimension_column.alias = context.alias(dimension)
            select.project(dimension_column)
            select.group_by(dimension_column.alias)

            self.sql_node_reference[dimension] = dimension_column.alias

    def _build_aggr_function_call_projection(self, context: SqlContext, select: Select,
                                             aggr_function_call_query: Select):
        for aggr_function_call in self._aggr_function_calls:
            aggr_function_call_column = SqlColumn(aggr_function_call_query,
                                                  self.right.get_reference(aggr_function_call))
            if isinstance(aggr_function_call, SqlAdditiveAggrFunctionCall):
                function_call_wrapper = SqlSumFunctionCall(aggr_function_call_column)
            else:
                function_call_wrapper = SqlAggFunctionCall(
                    aggr_function_call.function_name, aggr_function_call_column)

            function_call_wrapper.alias = context.alias(function_call_wrapper)
            select.project(function_call_wrapper)

            self.sql_node_reference[aggr_function_call] = function_call_wrapper.alias

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(context)


class DimensionQueryNode(SimpleQueryNode):

    def __init__(self, dimensions: List[SqlNode], function_call_primary_key: SqlColumn, query_scope: QueryScope):
        super().__init__(QueryNodeParameter())
        self._dimensions = dimensions
        self._function_call_primary_key = function_call_primary_key
        self._query_scope = query_scope

    def _rewrite_query_parameter(self, context: SqlContext, query_parameter: QueryNodeParameter):
        self._function_call_primary_key.alias = context.alias(self._function_call_primary_key)

        query_parameter.append_projections(self._function_call_primary_key)
        query_parameter.append_group_by_item(self._function_call_primary_key.alias)

        for dimension in self._dimensions:
            dimension.alias = context.alias(dimension)
            query_parameter.append_projections(dimension)
            query_parameter.append_group_by_item(dimension.alias)

            self.sql_node_reference[dimension] = dimension.alias

        self.sql_node_reference[self._function_call_primary_key] = self._function_call_primary_key.alias

        query_parameter.append_predicate_list(self._query_scope.where_predicates)

        return query_parameter

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(SqlContextWrapper(context))


class AdditiveAggrQueryNode(SimpleQueryNode):

    def __init__(self, additive_aggr_function_calls: List[SqlFunctionCall],
                 query_scope: QueryScope, function_call_primary_key: Optional[SqlColumn] = None):
        super().__init__(QueryNodeParameter())
        self._additive_aggr_function_calls = additive_aggr_function_calls
        self._function_call_primary_key = function_call_primary_key
        self._query_scope = query_scope

    def _rewrite_query_parameter(self, context: SqlContext, query_parameter: QueryNodeParameter):
        if self._function_call_primary_key is not None:
            self._function_call_primary_key.alias = context.alias(self._function_call_primary_key)
            query_parameter.append_projections(self._function_call_primary_key)
            query_parameter.append_group_by_item(self._function_call_primary_key.alias)

            self.sql_node_reference[self._function_call_primary_key] = self._function_call_primary_key.alias

        for additive_aggr_function_call in self._additive_aggr_function_calls:
            additive_aggr_function_call.alias = context.alias(additive_aggr_function_call)
            query_parameter.append_projections(additive_aggr_function_call)

            self.sql_node_reference[additive_aggr_function_call] = additive_aggr_function_call.alias

        query_scope = QueryScope.detect(context, [self._additive_aggr_function_calls])
        predicates = self._query_scope.get_predicates(query_scope.get_tables().values())
        query_parameter.append_predicate_list(predicates)

        return query_parameter

    def to_sql(self, context: 'SqlContext') -> str:
        return super().to_sql(SqlContextWrapper(context))


class SimpleAggrQueryNode(SimpleQueryNode):

    def __init__(self, aggr_function_calls: List[SqlAggFunctionCall],
                 query_scope: QueryScope, dimensions: Optional[List[SqlNode]] = None):
        super().__init__(QueryNodeParameter())
        self._aggr_function_calls = aggr_function_calls
        self._query_scope = query_scope
        self._dimensions = dimensions if dimensions is not None else []

    def _rewrite_query_parameter(self, context: SqlContext, query_parameter: QueryNodeParameter):
        for projection in (self._dimensions + self._aggr_function_calls):
            if hasattr(projection, "alias"):
                query_parameter.append_projections(projection)
                projection.alias = context.alias(projection)
                self.sql_node_reference[projection] = projection.alias

        for group_by_item in self._dimensions:
            if hasattr(group_by_item, "alias"):
                query_parameter.append_group_by_item(group_by_item.alias)

        for predicate in self._query_scope.where_predicates:
            query_parameter.append_predicate(predicate)

        return query_parameter


class QueryTree(SqlNode):

    def __init__(self, predicates: Optional[List[SqlPredicate]],
                 order_by_items: List[SqlNode],
                 limit: SqlLimit,
                 offset: SqlOffset):
        self._predicates = predicates
        self._order_by_items = order_by_items
        self._limit = limit
        self._offset = offset
        self._root = None

    def build_tree_nodes(self, context: SqlContext, projection_scope: ProjectionScope, query_scope: QueryScope):
        function_call_scope = projection_scope.aggr_function_call_scope
        if function_call_scope.has_additive_aggr_function_calls():
            if query_scope.has_multiple_tables():
                self._root = self._build_additive_aggr_query_tree(context, projection_scope, query_scope)
            else:
                self._root = self._create_simple_root(projection_scope, query_scope)
        else:
            self._root = self._create_simple_root(projection_scope, query_scope)

    def _build_additive_aggr_query_tree(self, context: SqlContext, projection_scope: ProjectionScope,
                                        query_scope: QueryScope, is_root: Optional[bool] = True) -> QueryNode:
        aggr_combine_nodes = self._create_additive_aggr_combine_nodes(
            context, projection_scope.dimensions, projection_scope.aggr_function_call_scope, query_scope)

        root = self._create_root_combine_query_node(projection_scope, is_root)
        parent = root
        for aggr_combine_node in aggr_combine_nodes:
            parent = self._append_combine_query_node(projection_scope, parent, aggr_combine_node)

        return root

    def _append_combine_query_node(self, projection_scope: ProjectionScope,
                                   parent: QueryNode, aggr_combine_node: QueryNode) -> QueryNode:
        if parent.left is None:
            parent.left = aggr_combine_node
        elif parent.right is None:
            parent.right = aggr_combine_node
        else:
            temp_parent = QueryCombineNode(projection_scope)
            temp_parent.left = parent.left
            temp_parent.right = aggr_combine_node
            parent.left = temp_parent
            parent = temp_parent
        return parent

    def _create_root_combine_query_node(self, projection_scope: ProjectionScope,
                                        is_root: Optional[bool] = True) -> QueryNode:
        return QueryCombineNode(
            projection_scope,
            order_by_items=self._order_by_items,
            limit=self._limit,
            offset=self._offset,
            is_root=is_root
        )

    def _create_additive_aggr_combine_nodes(self, context: SqlContext, dimensions: List[SqlNode],
                                            function_call_scope: FunctionCallScope,
                                            query_scope: QueryScope) -> List[QueryNode]:
        combine_nodes = list()
        simple_aggr_function_calls, tables = [], list(sorted(function_call_scope.function_calls.keys()))
        for table in tables:
            aggr_function_calls = function_call_scope.get_function_calls(table)
            if function_call_scope.has_additive_aggr_function_calls(table):
                primary_key = context.create_table_primary_key(table)
                combine_node = AdditiveAggrQueryCombineNode(dimensions, aggr_function_calls, primary_key)
                combine_node.left = DimensionQueryNode(dimensions, primary_key, query_scope)
                combine_node.right = AdditiveAggrQueryNode(aggr_function_calls, query_scope, primary_key)

                combine_nodes.append(combine_node)
            else:
                combine_nodes.append(SimpleAggrQueryNode(aggr_function_calls, query_scope, dimensions))

        return combine_nodes

    def _create_simple_root(self, projection_scope: ProjectionScope, query_scope: QueryScope) -> QueryNode:
        return self._create_simple_node(
            projections=projection_scope.projections,
            predicates=self._predicates,
            group_bys=(projection_scope.dimensions if query_scope.has_aggr_function_call() else []),
            is_root=True
        )

    def _create_simple_node(self, projections: Optional[List[SqlNode]] = None,
                            predicates: Optional[List[SqlNode]] = None,
                            group_bys: Optional[List[SqlNode]] = None,
                            is_root: Optional[bool] = False) -> QueryNode:
        return SimpleQueryNode(
            query_parameter=QueryNodeParameter(
                projections=projections,
                predicates=predicates,
                group_bys=group_bys,
                order_by_items=self._order_by_items,
                limit=self._limit,
                offset=self._offset
            ),
            is_root=is_root
        )

    def to_sql(self, context: 'SqlContext') -> str:
        return self._root.to_sql(context)


def create_query_tree(context: SqlContext, projections: List[SqlNode], predicates: Optional[List[SqlPredicate]] = None,
                      order_by_items: Optional[List[SqlOrderByItem]] = None, limit: Optional[SqlLimit] = None,
                      offset: Optional[SqlOffset] = None,
                      sql_predicate_appender: Optional[SqlPredicateAppender] = None):
    predicates = predicates if predicates is not None else []
    order_by_items = order_by_items if order_by_items is not None else []

    projection_scope = ProjectionScope.detect([projections])
    query_scope = QueryScope.detect(context, [projections, predicates, order_by_items])

    if sql_predicate_appender is not None:
        tables = sorted(query_scope.get_tables().values())
        for table in tables:
            sql_predicate_appender.append(query_scope, table)

    query_tree = QueryTree(query_scope.get_all_predicates(), order_by_items, limit, offset)
    query_tree.build_tree_nodes(context, projection_scope, query_scope)

    return query_tree
