from collections import defaultdict
from typing import List, Dict, Optional

from .algorithm import detect_involved_tables, decide_main_table
from .frozen import immutable_property
from .sql import SqlNode, SqlColumn, SqlAggFunctionCall, SqlNodeInspection, SqlFunctionCall, SqlTable, SqlLimit, \
    SqlOffset, SqlContext, SqlPredicate, \
    SqlAdditiveAggrFunctionCall


class QueryNodeParameter:

    def __init__(self,
                 projections: Optional[List[SqlNode]] = None,
                 predicates: Optional[List[SqlNode]] = None,
                 group_bys: Optional[List[SqlNode]] = None,
                 order_by_items: Optional[List[SqlNode]] = None,
                 limit: Optional[SqlLimit] = None,
                 offset: Optional[SqlOffset] = None):
        self._projections = projections if projections is not None else list()
        self._predicates = predicates if predicates is not None else list()
        self._group_bys = group_bys if group_bys is not None else list()
        self.order_by_items = order_by_items if order_by_items is not None else list()
        self.limit = limit
        self.offset = offset

    @immutable_property
    def projections(self):
        return self._projections

    @immutable_property
    def predicates(self):
        return self._predicates

    @immutable_property
    def group_bys(self):
        return self._group_bys

    @immutable_property
    def where_predicates(self) -> List[SqlNode]:
        predicates = []
        for predicate in self._predicates:
            if isinstance(predicate, SqlPredicate):
                if not isinstance(predicate, SqlAggFunctionCall):
                    predicates.append(predicate)
        return predicates

    @immutable_property
    def having_predicates(self) -> List[SqlNode]:
        predicates = []
        for predicate in self._predicates:
            if isinstance(predicate, SqlPredicate):
                if isinstance(predicate, SqlAggFunctionCall):
                    predicates.append(predicate)
        return predicates

    def append_projections(self, projection: SqlNode):
        if projection not in self._projections:
            self._projections.append(projection)

    def append_group_by_item(self, item: SqlNode):
        if item not in self._group_bys:
            self._group_bys.append(item)

    def append_predicate(self, predicate: SqlNode):
        if predicate not in self._predicates:
            self._predicates.append(predicate)

    def append_predicate_list(self, predicates: List[SqlNode]):
        for predicate in predicates:
            self.append_predicate(predicate)


class ProjectionScope:

    def __init__(self):
        self._projections = []
        self._dimensions = []
        self._aggr_function_calls = []
        self._additive_aggr_function_call = []
        self._function_call_scope = FunctionCallScope()

    @immutable_property
    def projections(self) -> List[SqlNode]:
        return self._projections

    @immutable_property
    def dimensions(self) -> List[SqlNode]:
        return self._dimensions

    @property
    def aggr_function_call_scope(self) -> 'FunctionCallScope':
        return self._function_call_scope

    def add_projection(self, projection: SqlNode):
        self._projections.append(projection)

        if isinstance(projection, SqlAggFunctionCall):
            table = QueryScope.simple_detect([projection]).get_single_table()
            self._function_call_scope.add_function_call(table, projection)
        else:
            self._dimensions.append(projection)

    def has_dimensions(self):
        return len(self._dimensions) > 0

    @staticmethod
    def detect(nodes_list: List[List[SqlNode]]) -> 'ProjectionScope':
        projection_scope = ProjectionScope()
        for nodes in nodes_list:
            for node in nodes:
                if isinstance(node, SqlNodeInspection):
                    node.inspect_projection(projection_scope)
        return projection_scope


class FunctionCallScope:

    def __init__(self):
        self._function_calls_dict = defaultdict(list)

    @immutable_property
    def function_calls(self) -> Dict[SqlTable, List[SqlFunctionCall]]:
        return self._function_calls_dict

    def get_function_calls(self, table: SqlTable) -> List[SqlFunctionCall]:
        return self._function_calls_dict.get(table)

    def has_additive_aggr_function_calls(self, table: Optional[SqlTable] = None) -> bool:
        function_call_list = self._function_calls_dict.values() if table is None \
            else [self._function_calls_dict.get(table)]
        for function_calls in function_call_list:
            for function_call in function_calls:
                if isinstance(function_call, SqlAdditiveAggrFunctionCall):
                    return True
        return False

    def has_aggr_function_calls(self, table: Optional[SqlTable] = None) -> bool:
        function_call_list = self._function_calls_dict.values() if table is None \
            else [self._function_calls_dict.get(table)]
        for function_calls in function_call_list:
            for function_call in function_calls:
                if isinstance(function_call, SqlAggFunctionCall):
                    return True
        return False

    def add_function_call(self, table: SqlTable, function_call: SqlFunctionCall):
        if table not in self._function_calls_dict or \
                function_call not in self._function_calls_dict[table]:
            self._function_calls_dict[table].append(function_call)


class QueryScope:

    def __init__(self):
        self._tables = {}
        self._columns = defaultdict(list)
        self._function_calls = []
        self._aggr_function_calls = []
        self._predicates_dict = defaultdict(list)

    @immutable_property
    def where_predicates(self) -> List[SqlNode]:
        predicates = []
        for _predicates in self._predicates_dict.values():
            for _predicate in _predicates:
                if not isinstance(_predicate, SqlAggFunctionCall):
                    predicates.append(_predicate)
        return predicates

    def add_column(self, column: SqlColumn):
        self.set_table(column.table)
        if isinstance(column.table, SqlTable):
            if column not in self._columns[column.table.name]:
                self._columns[column.table.name].append(column)
        else:
            raise RuntimeError(f"Column.table[{column}] must be a SqlTable")

    def add_predicate(self, table: SqlTable, predicate: SqlNode):
        if table not in self._predicates_dict or predicate not in self._predicates_dict.get(table):
            self._predicates_dict[table].append(predicate)

    def get_all_predicates(self):
        predicates = list()
        tables = sorted(list(self._predicates_dict.keys()))
        for table in tables:
            predicates = predicates + self._predicates_dict.get(table)
        return predicates

    def get_predicates(self, tables: List[SqlTable]) -> List[SqlNode]:
        predicates = list()
        for table in tables:
            _predicates = self._predicates_dict.get(table)
            if _predicates is not None:
                predicates = predicates + _predicates
        return predicates

    def add_function_call(self, function_call: SqlFunctionCall):
        self._function_calls.append(function_call)
        if isinstance(function_call, SqlAggFunctionCall):
            self._aggr_function_calls.append(function_call)

    def set_table(self, table: SqlTable):
        self._tables[table.name] = table

    def has_predicates(self):
        return len(self._predicates_dict) > 0

    def has_aggr_function_call(self):
        return len(self._aggr_function_calls) > 0

    def has_multiple_tables(self) -> bool:
        return len(self._tables) > 1

    def get_tables(self) -> Dict[str, SqlTable]:
        return self._tables

    def get_main_table(self, context: 'SqlContext') -> SqlTable:
        return decide_main_table(
            tables_name=sorted(list(self.get_tables().keys())),
            relation_graph=context.get_relation_graph()
        )

    def get_single_table(self) -> SqlTable:
        if len(self._tables) > 1:
            raise RuntimeError("Too many tables for single table query")
        return list(self._tables.values())[0]

    def _detect_involved_tables(self, context: 'SqlContext'):
        queried_tables_name = list(self._tables.keys())
        relation_graph = context.get_relation_graph()
        _, involved_tables_name = detect_involved_tables(
            queried_tables_name[0], queried_tables_name[1:], {queried_tables_name[0]}, relation_graph)
        for table_name in involved_tables_name:
            if not (table_name in self._tables):
                self._tables[table_name] = context.get_sql_table(table_name)
        return self._tables

    @staticmethod
    def detect(context: 'SqlContext', nodes_list: List[List[SqlNode]]) -> 'QueryScope':
        query_scope = QueryScope()
        for nodes in nodes_list:
            for node in nodes:
                if isinstance(node, SqlNodeInspection):
                    node.inspect(query_scope)
        if len(query_scope._tables) > 1:
            query_scope._detect_involved_tables(context)
        return query_scope

    @staticmethod
    def simple_detect(nodes: List[SqlNode]) -> 'QueryScope':
        query_scope = QueryScope()
        for node in nodes:
            if isinstance(node, SqlNodeInspection):
                node.inspect(query_scope)
        return query_scope
