from abc import ABCMeta, abstractmethod, ABC
from typing import Optional, List, Dict, Any

from flatql.schemas import Table


class SqlNodeInspection(metaclass=ABCMeta):

    @abstractmethod
    def inspect(self, scope: 'QueryScope'): pass

    def inspect_projection(self, projection_scope: 'ProjectionScope'): pass


class SqlNode(metaclass=ABCMeta):

    @abstractmethod
    def to_sql(self, context: 'SqlContext') -> str:
        pass

    def __hash__(self) -> int:
        return super().__hash__()

    def __eq__(self, other: Any) -> bool:
        return other.__class__ == self.__class__ and self.__dict__ == other.__dict__


class OverridableSqlNode(metaclass=ABCMeta):

    def __init__(self):
        self._original_alias = None

    @abstractmethod
    def override(self, overrode: SqlNode): pass

    @property
    def original_alias(self):
        return self._original_alias


class SqlKeyword(SqlNode):

    def __init__(self, keyword: str):
        self.keyword = keyword

    def to_sql(self, context: 'SqlContext') -> str:
        return self.keyword


class SqlOperator(SqlNode):

    def __init__(self, text: str):
        self.text = text

    def to_sql(self, context: 'SqlContext') -> str:
        return self.text


K_AND = SqlKeyword("AND")
K_OR = SqlKeyword("OR")
K_AS = SqlKeyword("AS")
K_ON = SqlKeyword("ON")

K_LEFT_JOIN = SqlKeyword("LEFT JOIN")
K_RIGHT_JOIN = SqlKeyword("RIGHT JOIN")
K_INNER_JOIN = SqlKeyword("INNER JOIN")
K_FULL_JOIN = SqlKeyword("FULL JOIN")
K_CROSS_JOIN = SqlKeyword("CROSS JOIN")

K_DISTINCT = SqlKeyword("DISTINCT")
K_ORDER_BY = SqlKeyword("ORDER BY")
K_ASC = SqlKeyword("ASC")
K_DESC = SqlKeyword("DESC")
K_LIMIT = SqlKeyword("LIMIT")
K_OFFSET = SqlKeyword("OFFSET")
K_HAVING = SqlKeyword("HAVING")

O_PLUS = SqlOperator("+")
O_MINUS = SqlOperator("-")
O_TIMES = SqlOperator("*")
O_DIVIDE = SqlOperator("/")
O_EQ = SqlOperator("=")
O_GT = SqlOperator(">")
O_LT = SqlOperator("<")
O_GE = SqlOperator(">=")
O_LE = SqlOperator("<=")
O_NE = SqlOperator("!=")

LPAREN = SqlOperator("(")
RPAREN = SqlOperator(")")


class SqlIdentifier(SqlNode):

    def __init__(self, identifier: str):
        self._identifier = identifier

    @property
    def value(self):
        return self._identifier

    def to_sql(self, context: 'SqlContext') -> str:
        from flatql.parser.ast import Identifier

        if isinstance(self._identifier, Identifier):
            return context.quote_identifier(self._identifier.value())
        return context.quote_identifier(self._identifier)


class SqlQualifiedName(SqlNode):

    def __init__(self, qualified_name: 'QualifiedName'):
        self.qualified_name = qualified_name

    def to_sql(self, context: 'SqlContext') -> str:
        return ".".join([context.quote_identifier(q.value()) for q in self.qualified_name.parts])


class SqlLiteral(SqlNode):

    def __init__(self, literal: 'Literal'):
        self.literal = literal

    def to_sql(self, context: 'SqlContext') -> str:
        from ..parser.ast import LITERAL_TYPE_STRING

        if self.literal.literal_type == LITERAL_TYPE_STRING:
            return context.quote_string_literal(self.literal.value())
        else:
            return self.literal.value()


class SqlContext(metaclass=ABCMeta):

    def __init__(self,
                 view: Optional['View'] = None,
                 tables_dict: Optional[Dict[str, 'SqlTable']] = None,
                 join_conditions: Optional[Dict[str, Dict[str, SqlNode]]] = None,
                 table_alias_prefix: Optional[str] = "ta_",
                 column_alias_prefix: Optional[str] = "co_",
                 subquery_alias_prefix: Optional[str] = "qu_"):
        self._view = view
        self._tables_dict = tables_dict
        self._join_conditions = join_conditions
        self.table_alias_prefix = table_alias_prefix
        self.column_alias_prefix = column_alias_prefix
        self.subquery_alias_prefix = subquery_alias_prefix
        self.table_list, self.column_list, self.subquery_list = [], [], []

    def alias(self, node: SqlNode) -> SqlIdentifier:
        from .select import Select

        if isinstance(node, SqlTable):
            if node not in self.table_list:
                self.table_list.append(node)
            return SqlIdentifier(f"{self.table_alias_prefix}{self.table_list.index(node)}")
        elif isinstance(node, Select):
            if node not in self.subquery_list:
                self.subquery_list.append(node)
            return SqlIdentifier(f"{self.subquery_alias_prefix}{self.subquery_list.index(node)}")
        else:
            if node not in self.column_list:
                self.column_list.append(node)
            return SqlIdentifier(f"{self.column_alias_prefix}{self.column_list.index(node)}")

    def create_table_primary_key(self, table: 'SqlTable'):
        primary_key = self.get_table_schema(table.name).primary_key
        return SqlColumn(table, SqlIdentifier(primary_key))

    def get_relation_graph(self):
        return self._view.get_relation_graph()

    def get_join_condition(self, parent_table_name: str, table_name: str) -> 'SqlNode':
        join_condition = self._join_conditions.get(parent_table_name).get(table_name)
        if join_condition is None:
            join_condition = self._join_conditions.get(table_name).get(parent_table_name)
        return join_condition

    def get_table_schema(self, name: str) -> Table:
        return self._view.get_table(name)

    def get_sql_table(self, name: str) -> 'SqlTable':
        return self._tables_dict.get(name)

    def quote_string_literal(self, string_literal: str) -> str:
        return quote_string_literal(string_literal)

    def quote_identifier(self, identifier: str) -> str:
        return double_quoted_identifier(identifier)


class SqlContextWrapper(SqlContext):

    def __init__(self, context: SqlContext):
        super().__init__(
            view=context._view,
            tables_dict=context._tables_dict,
            join_conditions=context._join_conditions
        )


class SqlColumn(SqlNode, SqlNodeInspection, OverridableSqlNode):

    def __init__(self, table: 'SqlNode', column_name: SqlIdentifier, name: Optional[str] = None,
                 alias: Optional[SqlIdentifier] = None) -> object:
        super(SqlColumn, self).__init__()
        self.table = table
        self.column_name = column_name
        self.name = name
        self._original_alias = alias
        self.alias = alias
        self._overrode = None

    def inspect(self, scope: 'QueryScope'):
        scope.add_column(self)

    def inspect_projection(self, projection_scope: 'ProjectionScope'):
        projection_scope.add_projection(self)

    def override(self, overrode: SqlNode):
        self._overrode = overrode

    def to_sql(self, context: SqlContext) -> str:
        if self._overrode is not None:
            return self._overrode.to_sql(context)
        else:
            table_alias = context.alias(self.table).to_sql(context)
            column_alias = "" if not self.alias else f" {K_AS.to_sql(context)} {self.alias.to_sql(context)}"
            return "".join([table_alias, ".", self.column_name.to_sql(context), column_alias])


class SqlTable(SqlNode):

    def __init__(self, name: str, table_name: SqlNode):
        self.name = name
        self.table_name = table_name

    def to_sql(self, context: SqlContext) -> str:
        table_alias = context.alias(self).to_sql(context)
        table_name = self.table_name.to_sql(context)
        return " ".join([table_name, K_AS.to_sql(context), table_alias])

    def __lt__(self, other):
        return self.name < other.name


class SqlPredicate(ABC, SqlNode, SqlNodeInspection):

    def __init__(self, operand: SqlNode, joiner: Optional[SqlKeyword] = K_AND):
        self.operand = operand
        self.joiner = joiner

    def inspect(self, scope: 'QueryScope'):
        if isinstance(self.operand, SqlNodeInspection):
            self.operand.inspect(scope)

        if isinstance(self.operand, SqlColumn):
            scope.add_predicate(self.operand.table, self)


class SqlComparison(SqlPredicate):

    def __init__(self, operand: SqlNode, operator: SqlOperator, rhs: SqlNode, joiner: Optional[SqlKeyword] = K_AND):
        super(SqlComparison, self).__init__(operand, joiner)
        self.operator = operator
        self.rhs = rhs

    def inspect(self, scope: 'QueryScope'):
        if isinstance(self.operand, SqlColumn):
            scope.add_predicate(self.operand.table, self)

        if isinstance(self.operand, SqlNodeInspection):
            self.operand.inspect(scope)

        if isinstance(self.rhs, SqlNodeInspection):
            self.rhs.inspect(scope)

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.operand.to_sql(context), self.operator.to_sql(context), self.rhs.to_sql(context)]
        return " ".join(sql)


class SqlIn(SqlPredicate):

    def __init__(self, operand: SqlNode, negated: bool, literals: List[SqlLiteral],
                 joiner: Optional[SqlKeyword] = K_AND):
        super(SqlIn, self).__init__(operand, joiner)
        self.negated = negated
        self.literals = literals

    def to_sql(self, context: SqlContext) -> str:
        literals = [li.to_sql(context) for li in self.literals]
        sql = [self.operand.to_sql(context), "NOT IN" if self.negated else "IN", f'({", ".join(literals)})']
        return " ".join(sql)


class SqlLike(SqlPredicate):

    def __init__(self, operand: SqlNode, negated: bool, string_literal: SqlLiteral,
                 joiner: Optional[SqlKeyword] = K_AND):
        super(SqlLike, self).__init__(operand, joiner)
        self.negated = negated
        self.string_literal = string_literal

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.operand.to_sql(context), "NOT LIKE" if self.negated else "LIKE",
               self.string_literal.to_sql(context)]
        return " ".join(sql)


class SqlIsNull(SqlPredicate):

    def __init__(self, operand: SqlNode, negated: bool, joiner: Optional[SqlKeyword] = K_AND):
        super(SqlIsNull, self).__init__(operand, joiner)
        self.negated = negated

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.operand.to_sql(context), "IS", "NOT NULL" if self.negated else "NULL"]
        return " ".join(sql)


class SqlBetween(SqlPredicate):

    def __init__(self, operand: SqlNode, negated: bool, literal1: SqlLiteral,
                 literal2: SqlLiteral, joiner: Optional[SqlKeyword] = K_AND):
        super(SqlBetween, self).__init__(operand, joiner)
        self.negated = negated
        self.literal1 = literal1
        self.literal2 = literal2

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.operand.to_sql(context), "NOT BETWEEN" if self.negated else "BETWEEN",
               self.literal1.to_sql(context), "AND", self.literal2.to_sql(context)]
        return " ".join(sql)


class SqlFunctionCall(SqlNode, SqlNodeInspection, OverridableSqlNode):

    def __init__(self, function_name: str, *args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlFunctionCall, self).__init__()
        self.function_name = function_name
        self.args = args
        self._original_alias = alias
        self.alias = alias
        self._overrode = None

    def inspect(self, scope: 'QueryScope'):
        scope.add_function_call(self)

        for arg in self.args:
            if isinstance(arg, SqlNodeInspection):
                arg.inspect(scope)

    def override(self, overrode: SqlNode):
        self._overrode = overrode

    def to_sql(self, context: SqlContext) -> str:
        if self._overrode is not None:
            return self._overrode.to_sql(context)
        else:
            args = [a.to_sql(context) for a in self.args]
            alias = f" {K_AS.to_sql(context)} {self.alias.to_sql(context)}" if self.alias is not None else ""
            sql = [self.function_name, f'({", ".join(args)})', alias]
            return "".join(sql)


class SqlNormalFunctionCall(SqlFunctionCall):

    def __init__(self, function_name: str, *args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlNormalFunctionCall, self).__init__(function_name, *args, alias=alias)

    def inspect_projection(self, projection_scope: 'ProjectionScope'):
        projection_scope.add_projection(self)


class SqlAggFunctionCall(SqlFunctionCall):

    def __init__(self, function_name: str, *args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlAggFunctionCall, self).__init__(function_name, *args, alias=alias)

    def inspect_projection(self, projection_scope: 'ProjectionScope'):
        projection_scope.add_projection(self)


class SqlCountDistinctFunctionCall(SqlAggFunctionCall):

    def __init__(self, *args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlCountDistinctFunctionCall, self).__init__("COUNT", *args, alias=alias)

    def to_sql(self, context: SqlContext) -> str:
        args = [a.to_sql(context) for a in self.args]
        alias = f" {K_AS.to_sql(context)} {self.alias.to_sql(context)}" if self.alias is not None else ""
        sql = [self.function_name, f'({K_DISTINCT} {", ".join(args)})', alias]
        return "".join(sql)

    def inspect_projection(self, projection_scope: 'ProjectionScope'):
        projection_scope.add_projection(self)


class SqlAdditiveAggrFunctionCall(SqlAggFunctionCall):

    def __init__(self, function_name: str, *args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlAdditiveAggrFunctionCall, self).__init__(function_name, *args, alias=alias)


class SqlSumFunctionCall(SqlAdditiveAggrFunctionCall):

    def __init__(self, *args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlSumFunctionCall, self).__init__("SUM", *args, alias=alias)

    def inspect_projection(self, projection_scope: 'ProjectionScope'):
        projection_scope.add_projection(self)


class SqlCountFunctionCall(SqlAdditiveAggrFunctionCall):

    def __init__(self, *args: List[SqlNode], alias: Optional[SqlNode] = None):
        super(SqlCountFunctionCall, self).__init__("COUNT", *args, alias=alias)

    def inspect_projection(self, projection_scope: 'ProjectionScope'):
        projection_scope.add_projection(self)


class Subquery(SqlNode):

    def __init__(self, select: SqlNode):
        self.select = select

    def to_sql(self, context: 'SqlContext') -> str:
        return " ".join([f"({self.select.to_sql(context)})", K_AS.to_sql(context),
                         context.alias(self.select).to_sql(context)])


class SqlLogicalExpression(SqlNode):

    def __init__(self, *predicates: List[SqlPredicate]):
        self.predicates = predicates

    def to_sql(self, context: 'SqlContext') -> str:
        return f" {K_AND.to_sql(context)} ".join([p.to_sql(context) for p in self.predicates])


class SqlJoin(SqlNode):

    def __init__(self, join_type: SqlKeyword, subclause: SqlNode, on_expr: Optional[SqlNode] = None):
        self.join_type = join_type
        self.subclause = subclause
        self.on_expr = on_expr

    def to_sql(self, context: SqlContext) -> str:
        sql = [self.join_type.to_sql(context), self.subclause.to_sql(context)]
        if self.on_expr is not None:
            sql.append(K_ON.to_sql(context))
            sql.append(self.on_expr.to_sql(context))
        return " ".join(sql)


class SqlOrderByItem(SqlNode, SqlNodeInspection):

    def __init__(self, item: SqlNode, direction: SqlKeyword):
        self.item = item
        self.direction = direction

    def inspect(self, scope: 'QueryScope'):
        if isinstance(self.item, SqlNodeInspection):
            self.item.inspect(scope)

    def to_sql(self, context: 'SqlContext') -> str:
        sql = [self.item.to_sql(context), self.direction.to_sql(context)]
        return " ".join(sql)


class SqlLimit(SqlNode):

    def __init__(self, limit: 'Limit'):
        self.limit = SqlLiteral(limit.literal)

    def to_sql(self, context: 'SqlContext') -> str:
        sql = [K_LIMIT.to_sql(context), self.limit.to_sql(context)]
        return " ".join(sql)


class SqlOffset(SqlNode):

    def __init__(self, offset: 'Offset'):
        self.offset = SqlLiteral(offset.literal)

    def to_sql(self, context: 'SqlContext') -> str:
        sql = [K_OFFSET.to_sql(context), self.offset.to_sql(context)]
        return " ".join(sql)


class SqlBinaryExpression(SqlNode, SqlNodeInspection):

    def __init__(self, lhs: SqlNode, operator: SqlNode, rhs: SqlNode):
        self.lhs = lhs
        self.operator = operator
        self.rhs = rhs

    def inspect(self, scope: 'QueryScope'):
        if isinstance(self.lhs, SqlNodeInspection):
            self.lhs.inspect(scope)

        if isinstance(self.rhs, SqlNodeInspection):
            self.rhs.inspect(scope)

    def inspect_projection(self, scope: 'ProjectionScope'):
        if isinstance(self.lhs, SqlNodeInspection):
            self.lhs.inspect_projection(scope)

        if isinstance(self.rhs, SqlNodeInspection):
            self.rhs.inspect_projection(scope)

    def to_sql(self, context: 'SqlContext') -> str:
        sql = [self.lhs.to_sql(context), self.operator.to_sql(context), self.rhs.to_sql(context)]
        return " ".join(sql)


class SqlParenExpression(SqlNode, SqlNodeInspection):

    def __init__(self, expression: SqlNode):
        self.expression = expression

    def inspect(self, scope: 'QueryScope'):
        if isinstance(self.expression, SqlNodeInspection):
            self.expression.inspect(scope)

    def inspect_projection(self, scope: 'ProjectionScope'):
        if isinstance(self.expression, SqlNodeInspection):
            self.expression.inspect_projection(scope)

    def to_sql(self, context: 'SqlContext') -> str:
        return f"({self.expression.to_sql(context)})"


def double_quoted_identifier(identifier: str) -> str:
    return f"\"{identifier}\""


def backquoted_identifier(identifier: str) -> str:
    return f"`{identifier}`"


def quote_string_literal(string_literal: str) -> str:
    return f"'{string_literal}'"
