from typing import List, Dict, Optional

from .algorithm import detect_involved_tables
from .query import create_query_tree, SqlPredicateAppender
from .sql import SqlIdentifier, SqlLiteral, SqlTable, SqlColumn, \
    SqlAggFunctionCall, SqlKeyword, SqlComparison, SqlOperator, SqlIn, SqlBetween, SqlLike, SqlIsNull, \
    SqlLimit, SqlOffset, SqlOrderByItem, SqlCountDistinctFunctionCall, SqlBinaryExpression, SqlParenExpression, SqlNode, \
    SqlLogicalExpression, SqlQualifiedName, SqlCountFunctionCall, SqlSumFunctionCall, SqlContext
from ..schemas import Schema


class SqlRewriter:

    def __init__(self, schema: Schema, sql_predicate_appender: Optional[SqlPredicateAppender] = None):
        self._view = None
        self._schema = schema
        self._tables_dict = {}
        self._projections = None
        self._predicates = None
        self._order_by_items = None
        self._limit = None
        self._offset = None
        self._sql_predicate_appender = sql_predicate_appender

    def set_view(self, view: 'View'):
        self._view = self._schema.get_view(view.name.value())

    def set_projections(self, projections: List['SqlNode']):
        self._projections = projections

    def set_predicates(self, predicates: List['SqlNode']):
        self._predicates = predicates

    def set_order_by_items(self, order_by_items: List[SqlOrderByItem]):
        self._order_by_items = order_by_items

    def set_limit(self, limit: SqlNode):
        self._limit = limit

    def set_offset(self, offset: SqlNode):
        self._offset = offset

    def rewrite_identifier(self, identifier: 'Identifier') -> 'SqlNode':
        return SqlIdentifier(identifier.value())

    def rewrite_literal(self, literal: 'Literal') -> 'SqlNode':
        return SqlLiteral(literal)

    def rewrite_column(self, column: 'Column') -> 'SqlNode':
        sql_table = column.table.rewrite(self)
        raw_column_name = column.name.rewrite(self)
        column_alias = column.alias.rewrite(self) if column.alias is not None else None

        table_schema = self._view.get_table(sql_table.name)
        db_column_name = table_schema.get_column(raw_column_name.value).db_column_name

        return SqlColumn(sql_table, SqlIdentifier(db_column_name), name=column.name.value(), alias=column_alias)

    def rewrite_function_call(self, function_call: 'FunctionCall') -> 'SqlNode':
        args = [a.rewrite(self) for a in function_call.args]
        alias = function_call.alias.rewrite(self) if function_call.alias is not None else None
        return rewrite_sql_function_call(function_call.name, args, alias)

    def rewrite_comparison(self, comparison: 'Comparison') -> 'SqlNode':
        sql_operand = comparison.operand.rewrite(self)
        sql_operator = SqlOperator(comparison.operator)
        sql_rhs = comparison.literal.rewrite(self)
        return SqlComparison(sql_operand, sql_operator, sql_rhs)

    def rewrite_in(self, in_: 'In') -> 'SqlNode':
        sql_operand = in_.operand.rewrite(self)
        literals = [a.rewrite(self) for a in in_.literals]
        return SqlIn(sql_operand, in_.negated, literals)

    def rewrite_between(self, between: 'Between') -> 'SqlNode':
        sql_operand = between.operand.rewrite(self)
        literal1 = between.literal1.rewrite(self)
        literal2 = between.literal2.rewrite(self)
        return SqlBetween(sql_operand, between.negated, literal1, literal2)

    def rewrite_like(self, like: 'Like') -> 'SqlNode':
        sql_operand = like.operand.rewrite(self)
        literal = like.string_literal.rewrite(self)
        return SqlLike(sql_operand, like.negated, literal)

    def rewrite_is_null(self, is_null: 'IsNull') -> 'SqlNode':
        sql_operand = is_null.operand.rewrite(self)
        return SqlIsNull(sql_operand, is_null.negated)

    def rewrite_order_by_item(self, order_by_item: 'OrderByItem') -> 'SqlNode':
        item = order_by_item.node.rewrite(self)
        direction = SqlKeyword(order_by_item.direction)
        return SqlOrderByItem(item, direction)

    def rewrite_limit(self, limit: 'Limit') -> 'SqlNode':
        return SqlLimit(limit)

    def rewrite_offset(self, offset: 'Offset') -> 'SqlNode':
        return SqlOffset(offset)

    def rewrite_table(self, table: 'Table') -> 'SqlNode':
        from flatql.parser.flatql_parser import parse_qualified_name
        if table.name.value() not in self._tables_dict:
            table_schema = self._view.get_table(table.name.value())
            self._tables_dict[table.name.value()] = SqlTable(table.name.value(), SqlQualifiedName(
                parse_qualified_name(table_schema.table_name)))

        return self._tables_dict[table.name.value()]

    def rewrite_binary_expression(self, binary_expression: 'BinaryExpression') -> 'SqlNode':
        lhs = binary_expression.lhs.rewrite(self)
        operator = SqlOperator(binary_expression.operator)
        rhs = binary_expression.rhs.rewrite(self)
        return SqlBinaryExpression(lhs, operator, rhs)

    def rewrite_paren_expression(self, paren_expression: 'ParenExpression') -> 'SqlNode':
        return SqlParenExpression(paren_expression.binary_expr.rewrite(self))

    def to_sql(self, context: Optional[SqlContext] = None) -> str:
        if context is None:
            context = self._create_sql_context()

        return create_query_tree(
            context=context,
            projections=self._projections,
            predicates=self._predicates,
            order_by_items=self._order_by_items,
            limit=self._limit,
            offset=self._offset,
            sql_predicate_appender=self._sql_predicate_appender
        ).to_sql(context)

    def rewrite_join_conditions(self) -> Dict[str, Dict[str, SqlNode]]:
        from flatql.parser.flatql_parser import parse_join_condition
        from ..schemas import Table

        join_conditions = dict()
        queried_tables_name = list(self._view.table_dict.keys())
        _, involved_tables_name = detect_involved_tables(
            queried_tables_name[0], queried_tables_name[1:], {queried_tables_name[0]}, self._view.get_relation_graph())
        for table_name in involved_tables_name:
            join_conditions[table_name] = dict()
            table: Table = self._view.get_table(table_name)
            for join in table.joins:
                if join.target in involved_tables_name:
                    predicates = parse_join_condition(join.on_expr)
                    join_condition = SqlLogicalExpression(*[p.rewrite(self) for p in predicates])
                    join_conditions[table.name][join.target] = join_condition

        return join_conditions

    def _create_sql_context(self):
        return SqlContext(
            view=self._view,
            tables_dict=self._tables_dict,
            join_conditions=self.rewrite_join_conditions()
        )


FUNC_NAME_SUM = "SUM"
FUNC_NAME_COUNT = "COUNT"
FUNC_NAME_COUNT_DISTINCT = "COUNT_DISTINCT"
AGG_FUNCTIONS = ["AVG", "MIN", "MAX"]


def rewrite_sql_function_call(function_name, args: List['SqlNode'], alias: SqlIdentifier):
    if function_name.upper() == FUNC_NAME_COUNT_DISTINCT:
        return SqlCountDistinctFunctionCall(*args, alias=alias)
    elif function_name.upper() == FUNC_NAME_COUNT:
        if len(args) == 1 and isinstance(args[0], SqlTable):
            return SqlCountFunctionCall(SqlKeyword("*"), alias=alias)
        return SqlCountFunctionCall(*args, alias=alias)
    elif function_name.upper() == FUNC_NAME_SUM:
        return SqlSumFunctionCall(*args, alias=alias)
    elif function_name.upper() in AGG_FUNCTIONS:
        return SqlAggFunctionCall(function_name.upper(), *args, alias=alias)
